@@ -54,14 +54,20 @@ from functools import partial, wraps
 from time import time
 from warnings import warn
 import argparse
+import cStringIO
 import inspect
 import linecache
 import os
 import re
 import runpy
+import shlex
 import sys
 import threading
 import zipfile
+try:
+    from IPython.core.magic import register_line_cell_magic
+except ImportError:
+    register_line_cell_magic = lambda x: x
 
 __all__ = (
     'ProfileBase', 'ProfileRunnerBase', 'Profile', 'ThreadProfile',
@@ -110,6 +116,9 @@ def _getFuncOrFile(func, module, line):
     else:
         return '%s:%s' % (func, line)
 
+def _isCallgrindName(filepath):
+    return os.path.basename(filepath).startswith('cachegrind.out.')
+
 class _FileTiming(object):
     """
     Accumulation of profiling statistics (line and call durations) for a given
@@ -294,7 +303,7 @@ class ProfileBase(object):
 
     def _getFileTiming(self, frame):
         try:
-            return self.global_dict[id(frame.f_globals)]
+            return self.global_dict[(id(frame.f_globals), frame.f_code.co_filename)]
         except KeyError:
             f_globals = frame.f_globals
             name = self._getFilename(frame.f_code.co_filename, f_globals)
@@ -407,11 +416,11 @@ class ProfileBase(object):
             files which may not match with profiled code.
         """
         print('version: 1', file=out)
-        if commandline is not None:
-            print('cmd:', commandline, file=out)
         print('creator: pprofile', file=out)
         print('event: usphit :us/hit', file=out)
         print('events: hits us usphit', file=out)
+        if commandline is not None:
+            print('cmd:', commandline, file=out)
         file_dict = self.file_dict
         if relative_path:
             convertPath = _relpath
@@ -548,8 +557,12 @@ class ProfileBase(object):
         """
         Similar to profile.Profile.dump_stats - but different output format !
         """
-        with _open(filename, 'w', errors='replace') as out:
-            self.annotate(out)
+        if _isCallgrindName(filename):
+            with open(filename, 'w') as out:
+                self.callgrind(out)
+        else:
+            with _open(filename, 'w', errors='replace') as out:
+                self.annotate(out)
 
     def print_stats(self):
         """
@@ -581,18 +594,18 @@ class ProfileRunnerBase(object):
             return func(*args, **kw)
 
     def runfile(self, fd, argv, fd_name='<unknown>', compile_flags=0,
-            dont_inherit=1):
+            dont_inherit=1, globals={}):
         with fd:
             code = compile(fd.read(), fd_name, 'exec', flags=compile_flags,
                 dont_inherit=dont_inherit)
         original_sys_argv = list(sys.argv)
+        ctx_globals = globals.copy()
+        ctx_globals['__file__'] = fd_name
+        ctx_globals['__name__'] = '__main__'
+        ctx_globals['__package__'] = None
         try:
             sys.argv[:] = argv
-            return self.runctx(code, {
-                '__file__': fd_name,
-                '__name__': '__main__',
-                '__package__': None,
-            }, None)
+            return self.runctx(code, ctx_globals, None)
         finally:
             sys.argv[:] = original_sys_argv
 
@@ -998,13 +1011,13 @@ def _relpath(name):
     """
     return os.path.normpath(os.path.splitdrive(name)[1]).lstrip(_allsep)
 
-def main():
+def _main(argv, stdin=None):
     format_dict = {
         'text': 'annotate',
         'callgrind': 'callgrind',
     }
 
-    parser = argparse.ArgumentParser()
+    parser = argparse.ArgumentParser(argv[0])
     parser.add_argument('script', help='Python script to execute (optionaly '
         'followed by its arguments)', nargs='?')
     parser.add_argument('argv', nargs=argparse.REMAINDER)
@@ -1025,7 +1038,7 @@ def main():
         'profiling when 0.')
     parser.add_argument('-m', dest='module',
         help='Searches sys.path for the named module and runs the '
-        'corresponding .py file as a script. When given, positional arguments'
+        'corresponding .py file as a script. When given, positional arguments '
         'become sys.argv[1:]')
 
     group = parser.add_argument_group(
@@ -1046,7 +1059,7 @@ def main():
         help='Include files whose name would have otherwise excluded. '
         'If no exclusion was specified, all paths are excluded first.')
 
-    options = parser.parse_args()
+    options = parser.parse_args(argv[1:])
     if options.exclude_syspath:
         options.exclude.extend('^' + re.escape(x) for x in sys.path)
     if options.include and not options.exclude:
@@ -1065,17 +1078,37 @@ def main():
         if options.script is None:
             parser.error('too few arguments')
         args = [options.script] + options.argv
-        runner_method_args = (args[0], args)
+        runner_method_kw = {
+            'path': args[0],
+            'argv': args,
+        }
         runner_method_id = 'runpath'
+    elif stdin is not None and options.module == '-':
+        # Undocumented way of using -m, used internaly by %%pprofile
+        args = ['<stdin>']
+        if options.script is not None:
+            args.append(options.script)
+        args.extend(options.argv)
+        import __main__
+        runner_method_kw = {
+            'fd': stdin,
+            'argv': args,
+            'fd_name': '<stdin>',
+            'globals': __main__.__dict__,
+        }
+        runner_method_id = 'runfile'
     else:
         args = [options.module]
         if options.script is not None:
             args.append(options.script)
         args.extend(options.argv)
-        runner_method_args = (options.module, args)
+        runner_method_kw = {
+            'module': options.module,
+            'argv': args,
+        }
         runner_method_id = 'runmodule'
     if options.format is None:
-        if os.path.basename(options.out).startswith('cachegrind.out.'):
+        if _isCallgrindName(options.out):
             options.format = 'callgrind'
         else:
             options.format = 'text'
@@ -1094,7 +1127,7 @@ def main():
             klass = Profile
         prof = runner = klass(verbose=options.verbose)
     try:
-        getattr(runner, runner_method_id)(*runner_method_args)
+        getattr(runner, runner_method_id)(**runner_method_kw)
     finally:
         if options.out == '-':
             out = _reopen(sys.stdout, errors='replace')
@@ -1146,5 +1179,24 @@ def main():
         # do not change exit status.
         sys.exit(1)
 
+def pprofile(line, cell=None):
+    """
+    Profile line execution.
+    """
+    if cell is None:
+        # TODO: detect and use arguments (statistical profiling, ...) ?
+        return run(line)
+    else:
+        return _main(['%%pprofile', '-m', '-'] + shlex.split(line), cStringIO.StringIO(cell))
+try:
+    register_line_cell_magic(pprofile)
+except Exception:
+    # ipython can be imported, but may not be currently running.
+    pass
+del pprofile
+
+def main():
+    _main(sys.argv)
+
 if __name__ == '__main__':
     main()
